/****************************************************************
 * Gimp Plugin for the Paint Shop Pro File Format
 *
 * not yet supports loading and saving of .psp files
 * (supported format is version 3.0 as in Paint Shop Pro 5.0)
 *
 * for details on the features supported/excluded in this release
 * see the documentation (either in this package or at
 * http://stud1.tuwien.ac.at/~e8826423/Linux/gimp-psp/
 *
 * Paint Shop Pro: Jasc Software, Inc.
 *
 * $Id: psp.c,v 1.4 1999/08/27 12:38:43 tl Exp $
 *
 * (c) 1999 Thomas Linder <e8826423@student.tuwien.ac.at>
 *
 * PSP File Format library based on the psp file-format specs:
 *   http://www.jasc.com/specs/psp.html
 *
 * this plugin was written with the help of the
 * plugin-doc by Kevin Turner
 * (http://www.poboxes.com/kevint/gimp/doc/plugin-doc.html)
 * and the source of the tiff and png plugins, and to some extent the
 * gif and psd plugins and the xcf file format routines.
 *
 * Inspired by the need to have a format for image interchange
 * between PSP and Gimp.
 ****************************************************************
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 ****************************************************************/

static char const rcsid[] = "$Id: psp.c,v 1.4 1999/08/27 12:38:43 tl Exp $";
static char const copyright[] =
    "Copyright (c) 1999  Thomas Linder, e8826423@student.tuwien.ac.at";    

#include "psp.h"

#include <stdio.h>

#if defined(HAS_ASPRINTF) && ! defined(__USE_GNU)
/* asprintf is only declared in stdio.h for GNU + (POSIX or not strict
   ANSI)... I hope to find a better way but I really don't like compiler
   warnings. */
extern int asprintf (char **, const char *, ...);
#endif

#include <stdlib.h>
#include <string.h>
#include <libpsp/psp.h>

#if PLUGIN_DEBUG
#include <unistd.h>
#endif

#include <gtk/gtk.h>
#include <libgimp/gimp.h>

#if VERBOSE
#warning *** Verbose messages enabled
#else
#warning --- Verbose messages disabled
#endif
    /* Default to GIMP 1.0.x */
#ifndef GIMP_MAJOR_VERSION
#define GIMP_MAJOR_VERSION 1
#endif
#ifndef GIMP_MINOR_VERSION
#define GIMP_MINOR_VERSION 0
#endif
#ifndef GIMP_MICRO_VERSION
#define GIMP_MICRO_VERSION 0
#endif

#if GIMP_MAJOR_VERSION >= 1 && GIMP_MINOR_VERSION >= 1 && GIMP_MICRO_VERSION >= 4
#define GIMP_HAVE_UNITS 1-1-4
#include <libgimp/gimpunit.h>
#endif

/* #define VERBOSE 0 */

    /* If you want to risk losing your file contents without being
     * warned, you may set this to zero
     */
#define AUTO_SAVE_DISABLED 1

#define USE_GIMP_PARASITES 0 /* should be GIMP_HAVE_PARASITES */


#define DEFAULT_DESCRIPTION "Created with The GIMP"

#define MIN(a, b) (((a) < (b)) ? (a) : (b))



typedef struct
{
    int	compression;
    
} PSPSaveVals;

PSPSaveVals pspvals =
{
    PSP_IMAGE_COMPR_NONE,
};

int psp_do_save = FALSE;

static void query (void);
static void run (char * name,
		 int nparams,
		 GParam *param,
		 int *nreturn_vals,
		 GParam **return_vals);
static gint32 load_psp_image (char * filename);
static gint save_psp_image (char * filename);
static gint save_dialog(void);

GPlugInInfo PLUG_IN_INFO =
{
    NULL,	/* init_proc */
    NULL,	/* quit_proc */
    query,	/* query_proc */
    run,	/* run_proc */
};

/*
static char * image_title = NULL;
static char * image_artist = NULL;
static char * image_copyright = NULL;
static char * image_description = NULL;
static char * image_creation = NULL;
static char * image_application = NULL;
*/

int gimp2psplayermode [] =
{
    PSP_LAYERMODE_NORMAL,
    PSP_LAYERMODE_DISSOLVE,
    PSP_LAYERMODE_MULTIPLY,
    PSP_LAYERMODE_SCREEN,
    PSP_LAYERMODE_OVERLAY,
    PSP_LAYERMODE_DIFFERENCE,
    -1,				/* Addition */
    -1,				/* Subtraction */
    PSP_LAYERMODE_DARKEN,
    PSP_LAYERMODE_LIGHTEN,
    PSP_LAYERMODE_HUE,
    PSP_LAYERMODE_SATURATION,
    PSP_LAYERMODE_COLOR,
    PSP_LAYERMODE_LUMINOSITY,
    PSP_LAYERMODE_DODGE,
};
int psp2gimplayermode [] =
{
    NORMAL_MODE,
    DARKEN_ONLY_MODE,
    LIGHTEN_ONLY_MODE,
    HUE_MODE,
    SATURATION_MODE,
    COLOR_MODE,
    VALUE_MODE,
    MULTIPLY_MODE,
    SCREEN_MODE,
    DISSOLVE_MODE,
    OVERLAY_MODE,
    -1,			/* Hard light not implemented in Gimp (1.0) */
    -1,			/* Soft light not implemented in Gimp (1.0) */
    DIFFERENCE_MODE,
#if GIMP_MAJOR_VERSION >= 1 && GIMP_MINOR_VERSION >= 1 && GIMP_MICRO_VERSION >= 3
    DIVIDE_MODE,	/* New in Gimp (1.1).  Be warned that Dodge is
			 * much different from Divide despite of what 
			 * the Gimp layermode selection says.
			 */
#else
    -1,			/* Dodge not implemented in Gimp (1.0) */
#endif
    MULTIPLY_MODE,
			/* Burn not implemented in Gimp (1.0)
    			 * Gimp 1.1 says this is MULTIPLY which I am
			 * VERY unsure about as this looks very
			 * different from Paint Shop Pro
			 */
    -1			/* Exclusion not implemented in Gimp (1.0) */
};

#if GIMP_HAVE_UNITS
struct psp2gimpunitstable
{
    GUnit unit;
    double factor;
};
struct psp2gimpunitstable psp2gimpunit[] = {
    { UNIT_PIXEL, 0.0 },
    { UNIT_INCH, 1.0 },
    { UNIT_MM, 2.54 },
};
#endif

/* gimp's plug-in main */
MAIN ()


/* This is used by gimp to get information about the procedures this
 * plug-in supports including all their parameters.
 */

static void
query ()
{
	/* Define the parameters */
    static GParamDef load_args[] =
    {
	{ PARAM_INT32, "run_mode", "Interactive, non-interactive" },
	{ PARAM_STRING, "filename", "The name of the file to load" },
	{ PARAM_STRING, "raw_filename", "The name of the file to load" },
    };

    static GParamDef load_return_vals[] =
    {
	{ PARAM_IMAGE, "image", "Output image" },
    };

    static GParamDef save_args[] =
    {
	{ PARAM_INT32, "run_mode", "Interactive, non-interactive" },
	{ PARAM_IMAGE, "image", "Image to save" },
	{ PARAM_DRAWABLE, "drawable", "Drawable to save" },
	{ PARAM_STRING, "filename", "The name of the file to save the "
	  "image in" },
	{ PARAM_STRING, "raw_filename", "The filename entered" },
	{ PARAM_INT32, "compression", "Compression type: { NONE (0), "
	  "Packbits (1)" },
    };

    static int nload_args = sizeof (load_args) / sizeof (load_args[0]);
    static int nload_return_vals = sizeof (load_return_vals)
	/ sizeof (load_return_vals[0]);

    static int nsave_args = sizeof (save_args) / sizeof (save_args[0]);

	/* Now tell gimp what we have here */
    gimp_install_procedure ("file_psp_load",
			    "load files of the Paint Shop Pro format.\n",
			    "The psp file format is Paint Shop Pro's native "
			    "format to store layered images in.  This "
			    "function can read these files to a certain "
			    "extent - only 24 bit layered images are "
			    "supported now, alpha channels, selections etc. "
			    "are silently ignored.\n"
			    "The psp-library (libpsp) is needed for this "
			    "function to work.\n"
			    "See http://stud1.tuwien.ac.at/~e8826423/Gimp/ "
			    "for more information.\n"
			    "Plugin version: $Revision: 1.4 $",
			    "Thomas Linder",
			    "Thomas Linder",
			    "$Date: 1999/08/27 12:38:43 $",
			    "<Load>/PSP",
			    NULL,
			    PROC_PLUG_IN,
			    nload_args, nload_return_vals,
			    load_args, load_return_vals);

    gimp_install_procedure ("file_psp_save",
			    "does not really save images to the Paint Shop "
			    "Pro format yet.\n",
			    "The psp file format is Paint Shop Pro's native "
			    "format to store layered images including alpha "
			    "channels, selections, ...  This "
			    "function will one day be able to create files "
			    "of this type at least to a certain extent."
			    "The psp-library (libpsp) is needed for this "
			    "function to work.\n"
			    "See http://stud1.tuwien.ac.at/~e8826423/Gimp/ "
			    "for more information.\n"
			    "Plugin version: $Revision: 1.4 $",
			    "Thomas Linder",
			    "Thomas Linder",
			    "$Date: 1999/08/27 12:38:43 $",
			    "<Save>/PSP",
			    "RGB*",
			    PROC_PLUG_IN,
			    nsave_args, 0,
			    save_args, NULL);

	/* register the file handlers */
    
	/* no prefixes, suffix .psp, string at first 32 bytes
	 * Note: the magig is not 100 percent conforming, but how do I
	 * represent the nullbyte in a string? Yes, of course I
	 * could... - but as long as I don't have more specific
	 * information this will have to do it (and supposedly will work
	 * quite nice.
	 */
    gimp_register_magic_load_handler ("file_psp_load", "psp", "",
				      "0,string,\"Paint Shop Pro Image "
				      "File\n\x1a\0\0");
    gimp_register_save_handler ("file_psp_save", "psp", "");

#if VERBOSE
    printf ("gimp-psp-plugin: interface initialized!\n");
#endif
}

#if PLUGIN_DEBUG
static void
waitforcontinue_close_callback (GtkWidget * widget,
				gpointer data)
{
    gtk_main_quit ();
}

static void
waitforcontinue_callback (GtkWidget * widget,
			  gpointer data)
{
    gtk_widget_destroy (GTK_WIDGET (data));
}

static void
waitforcontinue ()
{
    GtkWidget	*dialog,	/* Dialog window */
		*entry,
		*button;	/* OK/cancel buttons */
    gchar	**argv;		/* Fake command-line args */
    gint	argc;		/* Number of fake command-line args */
    char	pidtext[64];

    argc    = 2;
    argv    = g_new (gchar *, 1);
    argv[0] = g_strdup("pspsave");
    argv[1] = g_strdup("--gtk-debug");

    gtk_init (&argc, &argv);
    gtk_rc_parse (gimp_gtkrc());

	/* Open a dialog window... */

    dialog = gtk_dialog_new();
    gtk_window_set_title (GTK_WINDOW(dialog), "Press Continue to go on...");
    gtk_window_position (GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
    gtk_signal_connect (GTK_OBJECT(dialog), "destroy",
			(GtkSignalFunc) waitforcontinue_close_callback, NULL);

    button = gtk_button_new_with_label ("Continue");
    GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
    gtk_signal_connect (GTK_OBJECT(button), "clicked",
			(GtkSignalFunc) waitforcontinue_callback,
			GTK_OBJECT(dialog));
    gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dialog)->action_area),
			button, TRUE, TRUE, 0);
    gtk_widget_grab_default(button);
    gtk_widget_show(button);

    entry = gtk_entry_new ();

    sprintf (pidtext, "PSP plugin PID: %d", getpid ());

    gtk_entry_set_text (GTK_ENTRY (entry), pidtext);
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), entry,
			TRUE, TRUE, 0);
    gtk_widget_show (entry);
    
    gtk_widget_show (dialog);

    gtk_main();
    gdk_flush();
}
#endif PLUGIN_DEBUG

static void
run (char * name,
     int nparams,
     GParam * param,
     int * nreturn_vals,
     GParam ** return_vals)
{
    GParam * values;
    gint32 imageID;

#if USE_GIMP_PARASITES
	/* Don't lose the creator information */
	/* A lot of information is available in the psp file which
	 * should be preserved, the description shall be translated to
	 * the gimp-comment (or if it doesn't exists maybe use the title
	 * instead?)
	 * And there is also state information which fits better
	 * to the image than application
	 */
    Parasite * parasite;
#endif /* USE_GIMP_PARASITES */

#if PLUGIN_DEBUG
    waitforcontinue ();
#endif
    
    values = g_new (GParam, 2);
    *return_vals = values;
    *nreturn_vals = 1;

	/* Default to a failure */
    values[0].type          = PARAM_STATUS;
    values[0].data.d_status = STATUS_CALLING_ERROR;

#if VERBOSE
    printf ("\n**********\ngimp-psp-plugin (build %s %s): %s\n",
	    __DATE__, __TIME__, name);
#endif
    
    if (strcmp (name, "file_psp_load") == 0)
    {
	imageID = load_psp_image (param[1].data.d_string);

	if (imageID == -1)
	    values[0].data.d_status = STATUS_EXECUTION_ERROR;
	else
	{
	    values[1].type = PARAM_IMAGE;
	    values[1].data.d_image = imageID;
	    values[0].data.d_status = STATUS_SUCCESS;
	    *nreturn_vals = 2;
	}
    }
    else if (strcmp (name, "file_psp_save") == 0)
    {
	*nreturn_vals = 1;

#if USE_GIMP_PARASITES
	int image = param[1].data.d_int32;
	parasite = gimp_image_find_parasite (image, GIMP_PARASITE_COMMENT);
	if (parasite)
	    s = g_strdup (parasite->data);
	parasite_free (parasite);
	parasite = gimp_image_find_parasite (image, GIMP_PARASITE_TITLE);
	if (parasite)
	    s = g_strdup (parasite->data);
	parasite_free (parasite);
	parasite = gimp_image_find_parasite (image, GIMP_PARASITE_AUTHOR);
	if (parasite)
	    s = g_strdup (parasite->data);
	parasite_free (parasite);
	parasite = gimp_image_find_parasite (image, GIMP_PARASITE_COPYRIGHT);
	if (parasite)
	    s = g_strdup (parasite->data);
	parasite_free (parasite);
#endif /* USE_GIMP_PARASITES */
	switch (param[0].data.d_int32)
	{
	  case RUN_INTERACTIVE:
	      gimp_get_data ("file_psp_save", &pspvals);
	      if (!save_dialog ())
		  return;
	      break;
	  case RUN_NONINTERACTIVE:
	      g_message ("PSP Save disabled\n");
#if AUTO_SAVE_DISABLED
	      values[0].data.d_status = STATUS_CALLING_ERROR;
	      return;
#else
	      if (nparams < 6)
	      {
		  values[0].data.d_status = STATUS_CALLING_ERROR;
		  return;
	      }
	      else if (nparams == 6)
		  pspvals.compression = param[5].data.d_int32;

		  /* Only support uncompressed and packbits for now */
	      if ((pspvals.compression == PSP_IMAGE_COMPR_NONE)
		  || (pspvals.compression == PSP_IMAGE_COMPR_RLE))
	      {
		  values[0].data.d_status = STATUS_CALLING_ERROR;
		  return;
	      }
	      break;
#endif /* if AUTO_SAVE_DISABLED */

	  case RUN_WITH_LAST_VALS:
#if AUTO_SAVE_DISABLED
	      g_message ("PSP Save disabled\n");
	      values[0].data.d_status = STATUS_CALLING_ERROR;
	      return;
#else
	      gimp_get_data ("file_psp_save", &pspvals);
#if USE_GIMP_PARASITES
#endif /* USE_GIMP_PARASITES */
	      break;
#endif /* if AUTO_SAVE_DISABLED */
	  default:
	      break;
	}
	if (save_psp_image (param[3].data.d_string))
	{
		/* Save ok -> store data and return success */
	    gimp_set_data ("file_psp_save", &pspvals, sizeof (pspvals));
	    values[0].data.d_status = STATUS_SUCCESS;
	}
	else
	    values[0].data.d_status = STATUS_EXECUTION_ERROR;
    }
    return;
}

static int
set_channel (GPixelRgn * pixel_rgn,
	     unsigned long cols,
	     unsigned long rows,
	     int bpp,
	     guchar fill,
	     int offset,
	     int shift)
{
    unsigned long col;
    unsigned long row;
    unsigned long start;
    guchar * d;
    guint tile_height;    
    guchar * dest;
    
    tile_height = gimp_tile_height ();

    dest = g_new (guchar, tile_height * cols * bpp);
    
    for (start = 0, row = 0; row < rows; row++)
    {
	d = dest + offset + cols * (row % tile_height) * shift;

	if (row == start)
	{
#if 0
	    printf ("Call to get_rect at row %ld: extent: (%ld, %ld), "
		    "(%ld, %ld)\n",
		    row, 0, start, cols, MIN (tile_height, rows - row));
#endif
	    gimp_pixel_rgn_get_rect (pixel_rgn, dest, (int) 0, (int)
				     start, (int) cols, MIN
				     (tile_height, rows - row));
	}

	for (col = 0; col < cols; col++, d+=shift)
	{
	    *d = fill;
	}
	
	if ((((row + 1) % tile_height) == 0) || (row + 1 == rows))
	{
#if 0
	    printf ("Call to set_rect at row %ld: extent: (%ld, %ld), "
		    "(%ld, %ld)\n",
		    row, 0, start, cols, 1 + row - start);
#endif
	    gimp_pixel_rgn_set_rect (pixel_rgn, dest, (int) 0, (int)
				     start, (int) cols, (int) 1 + row - start);
	    start = row + 1;
	}
    }

    g_free (dest);

    return 1;
}


static int
load_channel (PSP * psp,
	      GPixelRgn * pixel_rgn,
	      unsigned long cols,
	      unsigned long rows,
	      int bpp,
	      int offset,
	      int shift)
{
    unsigned long col;
    unsigned long row;
    unsigned long start;
    guchar * data;
    guchar * s;
    guchar * d;
    guint tile_height;    
    guchar * dest;
    
    data = g_malloc (cols);

    tile_height = gimp_tile_height ();

    dest = g_new (guchar, tile_height * cols * bpp);
    
    for (start = 0, row = 0; row < rows; row++)
    {
	d = dest + offset + cols * (row % tile_height) * shift;
	s = data;

	if (row == start)
	{
#if 0
	    printf ("Call to get_rect at row %ld: extent: (%ld, %ld), "
		    "(%ld, %ld)\n",
		    row, 0, start, cols, MIN (tile_height, rows - row));
#endif
	    gimp_pixel_rgn_get_rect (pixel_rgn, dest, (int) 0, (int)
				     start, (int) cols, MIN
				     (tile_height, rows - row));
	}

	if (PSPReadScanline (psp, data, row))
	{
#if 0
	    {
		unsigned long i;
		
		if (row == 0)
		{
		    printf ("first few bytes:");
		    for (i = 0; i < MIN (cols, 20); i++)
		    {
			printf (" %02x", data[i]);
		    }
		    printf ("\n");
		} else if (row == rows -1)
		{
		    printf ("last few bytes: ");
		    for (i = col - MIN (cols, 20); i < cols; i++)
		    {
			printf (" %02x", data[i]);
		    }
		    printf ("\n");
		}
	    }
#endif
	    for (col = 0; col < cols; col++, d+=shift, s++)
	    {
		*d = *s;
	    }
	}
	else
	{
	    g_free (data);
	    g_free (dest);
	    return 0;
	}
	
	if ((((row + 1) % tile_height) == 0) || (row + 1 == rows))
	{
#if 0
	    printf ("Call to set_rect at row %ld: extent: (%ld, %ld), "
		    "(%ld, %ld)\n",
		    row, 0, start, cols, 1 + row - start);
#endif
	    gimp_pixel_rgn_set_rect (pixel_rgn, dest, (int) 0, (int)
				     start, (int) cols, (int) 1 + row - start);
	    start = row + 1;
	}
    }

    g_free (data);
    g_free (dest);

    return 1;
}

static gint32
load_layer (PSP * psp,
	    gint32 image,
	    GImageType image_type)
{
    pspbyte_t pspopacity;
    gdouble opacity;
    int transparencymask;
    gint32 layer;
    GDrawable * drawable;
    GDrawableType drawable_type;
    GPixelRgn pixel_rgn;
    GLayerMode layer_mode;
    int visible;
    int blendingmode;
    int transprot;
    char * name;
#if HAS_ASPRINTF
    char * newname;
#else
    char newname [256];
#endif
    int shift, transoffset;
    psprect_t extent;
    psprect_t dataextent;
    psprect_t maskextent;
    unsigned long rows, cols;
    unsigned long maskrows, maskcols;
    pspcht_t chtype;

    if (!PSPGetField (psp, PSPF_LAYERNAME, &name)
	|| !PSPGetField (psp, PSPF_LAYEROPACITY, &pspopacity)
	|| !PSPGetField (psp, PSPF_LAYERMODE, &blendingmode)
	|| !PSPGetField (psp, PSPF_LAYERVISIBLE, &visible)
	|| !PSPGetField (psp, PSPF_LAYERTRANSPROT, &transprot)
	)
    {
	g_message ("PSP Load: Unable to read layer attributes");
	return -1;
    }
    opacity = ((gdouble) pspopacity * 100 / 255);

    transparencymask = PSPSetChannel (psp, PSP_CHANNELTRANSMASK);

    if (image_type == RGB)
    {
	if (transparencymask)
	    drawable_type = RGBA_IMAGE;
	else
	    drawable_type = RGB_IMAGE;
	shift = 3;
	transoffset = 3;
    }
    else if (image_type == GRAY)
    {
	if (transparencymask)
	    drawable_type = GRAYA_IMAGE;
	else
	    drawable_type = GRAY_IMAGE;
	shift = 1;
	transoffset = 1;
    }
    else
    {
	if (transparencymask)
	    drawable_type = INDEXEDA_IMAGE;
	else
	    drawable_type = INDEXED_IMAGE;
	shift = 1;
	transoffset = 1;
    }
    if (transparencymask)
	shift++;

	    /* If the blending mode does not exist, set the layer
	       invisible. */
    if ((layer_mode = psp2gimplayermode[blendingmode]) == -1)
    {
	layer_mode = NORMAL_MODE;
	visible = FALSE;
#if VERBOSE
	g_message ("PSP Load: Layer blending mode not supported!");
#endif
	switch (blendingmode)
	{
	  case PSP_LAYERMODE_HARDLIGHT:
#if defined (HAS_ASPRINTF)
	      asprintf (&newname, "*** OFF: Hard Light - %s", name);
#else
	      sprintf (newname, "*** OFF: Hard Light - %.200s", name);
#endif
	      break;
	  case PSP_LAYERMODE_SOFTLIGHT:
#if defined (HAS_ASPRINTF)
	      asprintf (&newname, "*** OFF: Soft Light - %s", name);
#else
	      asprintf (&newname, "*** OFF: Soft Light - %s", name);
#endif
	      break;
	  case PSP_LAYERMODE_DODGE:
#if defined (HAS_ASPRINTF)
	      asprintf (&newname, "*** OFF: Dodge - %s", name);
#else
	      asprintf (&newname, "*** OFF: Dodge - %s", name);
#endif
	      break;
	  case PSP_LAYERMODE_BURN:
#if defined (HAS_ASPRINTF)
	      asprintf (&newname, "*** OFF: Burn - %s", name);
#else
	      asprintf (&newname, "*** OFF: Burn - %s", name);
#endif
	      break;
	  case PSP_LAYERMODE_EXCLUSION:
#if defined (HAS_ASPRINTF)
	      asprintf (&newname, "*** OFF: Exclusion - %s", name);
#else
	      asprintf (&newname, "*** OFF: Exclusion - %s", name);
#endif
	      break;
	  default:
#if defined (HAS_ASPRINTF)
	      asprintf (&newname, "*** OFF: Unknown mode - %s", name);
#else
	      asprintf (&newname, "*** OFF: Unknown mode - %s", name);
#endif
	}
    }
    else
	newname = name;

    if (PSPGetField (psp, PSPF_LAYERDATAEXTENT, &dataextent))
    {
	cols = dataextent.xmax - dataextent.xmin;
	rows = dataextent.ymax - dataextent.ymin;
    }
    else
    {
	g_message ("PSP Load: Read failed\n");
	return -1;
    }
	
    if (!PSPGetField (psp, PSPF_LAYEREXTENT, &extent))
    {
	g_message ("PSP Load: Read failed\n");
	return -1;
    }

	/* FIXME: Not sure what to do with it, offsets may
	     * not be correct when size differs from normal extent
	     */
    if (PSPGetField (psp, PSPF_LAYERMASKDATAEXTENT, &maskextent))
    {
	maskcols = maskextent.xmax - maskextent.xmin;
	maskrows = maskextent.ymax - maskextent.ymin;
    }
    else
    {
	g_message ("PSP Load: Read failed\n");
	return -1;
    }

#if VERBOSE
    printf ("Layermode: %d\n", layer_mode);
    printf ("gimp_layer_new (image = %d,\n", image);
    printf ("                newname = %s,\n", newname);
    printf ("                cols = %ld,\n", cols);
    printf ("                rows = %ld,\n", rows);
    printf ("                drawable_type = %d,\n", image);
    printf ("                opacity = %.2f,\n", opacity);
    printf ("                layer_mode = %d);\n", layer_mode);
	
#endif
#if ALTERNATE_MODE_METHOD
    layer = gimp_layer_new (image, newname, cols, rows,
			    drawable_type, opacity, NORMAL_MODE);
#else
    layer = gimp_layer_new (image, newname, cols, rows,
			    drawable_type, opacity, layer_mode);
#endif
    PSPFree (name);

	    /* If layer creation fails continue with the next layer
	     * and send a notice
	     */
    if (layer == -1)
    {
	g_message ("PSP Load: Layer creation failed. Image not complete!");
	return -1;
    }

#if ALTERNATE_MODE_METHOD
    gimp_layer_set_mode (layer, layer_mode);
#endif
	
#if VERBOSE
    printf ("Layer: %d\n", layer);
    printf ("gimp_layer_set_offsets (layer = %d,\n", layer);
    printf ("                        offx = %ld,\n",
	    extent.xmin + dataextent.xmin);
    printf ("                        offy = %ld,\n",
	    extent.ymin + dataextent.ymin);
#endif

    gimp_layer_set_visible (layer, visible);
    gimp_layer_set_preserve_transparency (layer, transprot);

    gimp_image_add_layer (image, layer, 0);

    gimp_layer_set_offsets (layer, (gint32) extent.xmin + dataextent.xmin,
			    (gint32) extent.ymin + dataextent.ymin);
    
    drawable = gimp_drawable_get (layer);

#if VERBOSE
    printf ("Drawable: 0x%08lux\n", (unsigned long) drawable);
    printf ("Drawable bpp: %d\n", drawable->bpp);
#endif

    gimp_pixel_rgn_init (&pixel_rgn, drawable, 0, 0,
			 cols, rows, TRUE, FALSE);

    if (!PSPSetChannel (psp, PSP_CHANNELFIRST))
    {
	g_message ("PSP Load Error: Unable to select channel\n");
	return -1;
    }

    do {
	if (!PSPGetField (psp, PSPF_CHANNELTYPE, &chtype))
	{
	    g_message ("PSP Load Error: Unable to get channel type\n");
	    return -1;
	}
	switch (chtype)
	{
	  case PSP_CHANNELRED:
	      if (!load_channel (psp, &pixel_rgn, cols, rows,
				 drawable->bpp, 0, shift))
	      {
		  g_message ("PSP Load Error: Read of channel content "
			     "failed\n");
		  return -1;
	      }
	      break;
	  case PSP_CHANNELGREEN:
	      if (!load_channel (psp, &pixel_rgn, cols, rows,
				 drawable->bpp, 1, shift))
	      {
		  g_message ("PSP Load Error: Read of channel content "
			     "failed\n");
		  return -1;
	      }
		  
	      break;
	  case PSP_CHANNELBLUE:
	      if (!load_channel (psp, &pixel_rgn, cols, rows,
				 drawable->bpp, 2, shift))
	      {
		  g_message ("PSP Load Error: Read of channel content "
			     "failed\n");
		  return -1;
	      }
	      break;
	  case PSP_CHANNELTRANSMASK:
	      if (!load_channel (psp, &pixel_rgn, cols, rows,
				 drawable->bpp, transoffset, shift))
	      {
		  g_message ("PSP Load Error: Read of channel content "
			     "failed\n");
		  return -1;
	      }
	      break;
	  case PSP_CHANNELCOMPOSITE:
	      if (!load_channel (psp, &pixel_rgn, cols, rows,
				 drawable->bpp, 0, shift))
	      {
		  g_message ("PSP Load Error: Read of channel content "
			     "failed\n");
		  return -1;
	      }
	      break;
	  case PSP_CHANNELUSERMASK:
	  {
	      gint32 mask;
	      GDrawable * drawable;
	      GPixelRgn pixel_rgn;
	      pspbool_t disabled;
		  
		  /* UNKNOWN: should it really be 0? */
	      mask = gimp_layer_create_mask (layer, 0);
	      gimp_image_add_layer_mask (image, layer, mask);

	      drawable = gimp_drawable_get (mask);
	      gimp_pixel_rgn_init (&pixel_rgn, drawable, 0, 0,
				   cols, rows, TRUE, FALSE);
	      if (!load_channel (psp, &pixel_rgn, maskcols, maskrows,
				 drawable->bpp, 0, 1))
	      {
		  g_message ("PSP Load Error: Read of selection content "
			     "failed\n");
		  return -1;
	      }
	      gimp_drawable_flush (drawable);
	      gimp_drawable_detach (drawable);

	      if (PSPGetField (psp, PSPF_LAYERMASKDISABLED, &disabled))
		  gimp_layer_set_apply_mask (layer,
					     (gint32) (disabled == FALSE));
	      gimp_layer_set_edit_mask (layer, 0);	      
	      break;
	  }
	  default:
#if VERBOSE
	      g_message ("Unsupported channel type in layer\n");
#endif
	}
    } while (PSPNextChannel(psp));

    gimp_drawable_flush (drawable);
    gimp_drawable_detach (drawable);

    return layer;
}

    /* Interface changed */
#if (GIMP_MAJOR_VERSION == 1) && (GIMP_MINOR_VERSION == 0)
static gint
gimp_channel_ops_offset (gint32 imageID,
			 gint32 drawableID,
			 gint32 wrap_around,
			 gint32 fill_type,
			 gint32 offset_x,
			 gint32 offset_y)
{
    GParam * return_vals;
    int nreturn_vals;
    gint ret;

#if VERBOSE
    printf ("gimp_channel_ops_offset (gint32 imageID = %d,\n"
	    "                         gint32 drawableID = %d,\n"
	    "                         gint32 wrap_around = %d,\n"
	    "                         gint32 fill_type = %d,\n"
	    "                         gint32 offset_x = %d,\n"
	    "                         gint32 offset_y = %d);\n",
	    imageID, drawableID, wrap_around, fill_type, offset_x, offset_y);
#endif
    
    return_vals =
	gimp_run_procedure ("gimp_channel_ops_offset",
			    &nreturn_vals,
			    PARAM_IMAGE, imageID,
			    PARAM_DRAWABLE, drawableID,
			    PARAM_INT32, wrap_around,
			    PARAM_INT32, fill_type,
			    PARAM_INT32, offset_x,
			    PARAM_INT32, offset_y,
			    PARAM_END);
#if VERBOSE
    printf ("gimp_channel_ops_offset: Number of return values: %d\n",
	    nreturn_vals);
    if (nreturn_vals > 0)
	printf ("  . Status: %s\n",
		(return_vals[0].data.d_status == STATUS_SUCCESS) ?
		"Success" : "Failure");
#endif
    
    ret = return_vals[0].data.d_status == STATUS_SUCCESS;
    
    gimp_destroy_params (return_vals, nreturn_vals);

    return ret;
}
#else /* if Gimp version = 1.0.x */
static gint
gimp_channel_ops_offset (gint32 drawableID,
			 gint32 wrap_around,
			 gint32 fill_type,
			 gint32 offset_x,
			 gint32 offset_y)
{
    GParam * return_vals;
    int nreturn_vals;
    gint ret;

#if VERBOSE
    printf ("gimp_channel_ops_offset (gint32 drawableID = %d,\n"
	    "                         gint32 wrap_around = %d,\n"
	    "                         gint32 fill_type = %d,\n"
	    "                         gint32 offset_x = %d,\n"
	    "                         gint32 offset_y = %d);\n",
	    drawableID, wrap_around, fill_type, offset_x, offset_y);
#endif
    
    return_vals =
	gimp_run_procedure ("gimp_channel_ops_offset",
			    &nreturn_vals,
			    PARAM_DRAWABLE, drawableID,
			    PARAM_INT32, wrap_around,
			    PARAM_INT32, fill_type,
			    PARAM_INT32, offset_x,
			    PARAM_INT32, offset_y,
			    PARAM_END);
#if VERBOSE
    printf ("gimp_channel_ops_offset: Number of return values: %d\n",
	    nreturn_vals);
    if (nreturn_vals > 0)
	printf ("  . Status: %s\n",
		(return_vals[0].data.d_status == STATUS_SUCCESS) ?
		"Success" : "Failure");
#endif
    
    ret = return_vals[0].data.d_status == STATUS_SUCCESS;
    
    gimp_destroy_params (return_vals, nreturn_vals);

    return ret;
}
#endif /* if Gimp version = 1.0.x */

#if 0
static gint
gimp_edit_clear (gint32 drawableID)
{
    GParam * return_vals;
    int nreturn_vals;
    gint ret;

    return_vals =
	gimp_run_procedure ("gimp_edit_clear",
			    &nreturn_vals,
			    PARAM_DRAWABLE, drawableID,
			    PARAM_END);
#if VERBOSE
    printf ("gimp_edit_clear: Number of return values: %d\n", nreturn_vals);
    if (nreturn_vals > 0)
	printf ("  . Status: %s\n",
		(return_vals[0].data.d_status == STATUS_SUCCESS) ?
		"Success" : "Failure");
#endif
    
    ret = return_vals[0].data.d_status == STATUS_SUCCESS;
    
    gimp_destroy_params (return_vals, nreturn_vals);

    return ret;
}
#endif

#if (GIMP_MAJOR_VERSION == 1) && (GIMP_MINOR_VERSION == 0)
    /* Interface changed */
static gint
gimp_selection_load (gint32 imageID,
		     gint32 channelID)
{
    GParam * return_vals;
    int nreturn_vals;
    gint ret;

#if VERBOSE
    printf ("gimp_selection_load (gint32 imageID = %d,\n"
	    "                     gint32 channelID = %d);\n",
	    imageID, channelID);
#endif
    
    return_vals =
	gimp_run_procedure ("gimp_selection_load",
			    &nreturn_vals,
			    PARAM_IMAGE, imageID,
			    PARAM_CHANNEL, channelID,
			    PARAM_END);
#if VERBOSE
    printf ("gimp_selection_load: Number of return values: %d\n",
	    nreturn_vals);
    if (nreturn_vals > 0)
	printf ("  . Status: %s\n",
		(return_vals[0].data.d_status == STATUS_SUCCESS) ?
		"Success" : "Failure");
#endif
    
    ret = return_vals[0].data.d_status == STATUS_SUCCESS;
    
    gimp_destroy_params (return_vals, nreturn_vals);

    return ret;
}
#else /* if Gimp version = 1.0.x */
static gint
gimp_selection_load (gint32 channelID)
{
    GParam * return_vals;
    int nreturn_vals;
    gint ret;

#if VERBOSE
    printf ("gimp_selection_load (gint32 channelID = %d);\n",
	    channelID);
#endif
    
    return_vals =
	gimp_run_procedure ("gimp_selection_load",
			    &nreturn_vals,
			    PARAM_CHANNEL, channelID,
			    PARAM_END);
#if VERBOSE
    printf ("gimp_selection_load: Number of return values: %d\n",
	    nreturn_vals);
    if (nreturn_vals > 0)
	printf ("  . Status: %s\n",
		(return_vals[0].data.d_status == STATUS_SUCCESS) ?
		"Success" : "Failure");
#endif
    
    ret = return_vals[0].data.d_status == STATUS_SUCCESS;
    
    gimp_destroy_params (return_vals, nreturn_vals);

    return ret;
}
#endif /* if Gimp version = 1.0.x */

#if GIMP_MAJOR_VERSION >= 1 && GIMP_MINOR_VERSION >= 1
static gint
gimp_layer_set_linked (gint32 layerID,
		       gint32 linkstate)
{
    GParam * return_vals;
    int nreturn_vals;
    gint ret;

#if VERBOSE
    printf ("gimp_layer_set_linked (gint32 layerID = %d,"
	    "                       gint32 link state = %d);\n",
	    layerID, linkstate);
#endif

    return_vals =
	gimp_run_procedure ("gimp_layer_set_linked",
			    &nreturn_vals,
			    PARAM_LAYER, layerID,
			    PARAM_INT32, linkstate,
			    PARAM_END);
#if VERBOSE
    printf ("gimp_layer_set_linked: Number of return values: %d\n",
	    nreturn_vals);
    if (nreturn_vals > 0)
	printf ("  . Status: %s\n",
		(return_vals[0].data.d_status == STATUS_SUCCESS) ?
		"Success" : "Failure");
#endif
    
    ret = return_vals[0].data.d_status == STATUS_SUCCESS;
    
    gimp_destroy_params (return_vals, nreturn_vals);

    return ret;
}
#endif /* GIMP_VERSION >= 1.1.0 */

/* Unknown to me if this is an exact distinction */
#if GIMP_MAJOR_VERSION >= 1 && GIMP_MINOR_VERSION >= 1
static int
gimp_floating_sel_attach (const gint32 layerID,
			  const gint32 drawableID)
{
    GParam *return_vals;
    int nreturn_vals;
    int ret;
    

#if VERBOSE
    printf ("gimp_floating_sel_attach (const gint32 = %d,", layerID);
    printf ("                          const gint32 = %d) = ", drawableID);
#endif

    return_vals = gimp_run_procedure ("gimp_floating_sel_attach",
				      &nreturn_vals,
				      PARAM_LAYER, layerID,
				      PARAM_DRAWABLE, drawableID,
				      PARAM_END);

    ret = return_vals[0].data.d_status == STATUS_SUCCESS;

#if VERBOSE
    printf ("%s\n", ret ? "Yes" : "No");
#endif
    
    gimp_destroy_params (return_vals, nreturn_vals);

    return ret;
}
#endif /* GIMP_VERSION >= 1.1.0 */

#if 0
static int
gimp_procedure_exists (const char * proc_name)
{
    GParam *return_vals;
    int nreturn_vals;
    int ret;
    

#if VERBOSE
    printf ("gimp_procedure_exists (const char * = %s) = ",
	    proc_name);
#endif

    return_vals = gimp_run_procedure ("gimp_procedural_db_proc_info",
				      &nreturn_vals,
				      PARAM_STRING, proc_name,
				      PARAM_END);

    ret = return_vals[0].data.d_status == STATUS_SUCCESS;

#if VERBOSE
    printf ("%s\n", ret ? "Yes" : "No");
#endif
    
    gimp_destroy_params (return_vals, nreturn_vals);

    return ret;
}
#endif

static gint32
load_psp_image (char * filename) 
{
    PSP * psp;
    char * name;
    gint32 imageID;
    int i;
    long width, height;
    GImageType image_type;
    pspword_t bpp;
    int greyscale;
    long layercount;
    long activelayer;
    gint32 layerID;
    gint32 activelayerID = -1;
    psplay_t layernum;
    guint tile_height;
    pspenum_t compression;
#if GIMP_MAJOR_VERSION >= 1 && GIMP_MINOR_VERSION >= 1
    pspword_t linkgroup;
    pspword_t firstlinkgroup;
#endif /* GIMP_VERSION >= 1.1.0 */
#if GIMP_HAVE_UNITS
    GUnit unit;
    double resolution;
#endif
    
    if ((psp = PSPOpen (filename, "rb")) == NULL)
    {
	g_message ("PSP Load: Not a PSP file");
	return -1;
    }

    name = g_malloc (strlen (filename) + 22);
    sprintf (name, "Loading PSP Image %s:", filename);
    gimp_progress_init (name);
    g_free (name);
    
    if (!PSPGetField (psp, PSPF_IMAGECOMPRTYPE, &compression)
	|| !PSPGetField (psp, PSPF_IMAGEWIDTH, &width)
	|| !PSPGetField (psp, PSPF_IMAGEHEIGHT, &height)
	|| !PSPGetField (psp, PSPF_IMAGEBITDEPTH, &bpp)
	|| !PSPGetField (psp, PSPF_IMAGEISGREYSCALE, &greyscale)
	|| !PSPGetField (psp, PSPF_IMAGELAYERCOUNT, &layercount))
    {
	g_message ("PSP Load: Unable to read image attributes");
	return -1;
    }

    if (compression == PSP_IMAGE_COMPR_LZ77)
    {
	g_message ("PSP Load: LZ77 compression not supported");
	return -1;
    }
    
    if (bpp == 24)
	image_type = RGB;
    else if (greyscale)
	image_type = GRAY;
    else
	image_type = INDEXED;

    if ((imageID = gimp_image_new (width, height, image_type)) == -1)
    {
	g_message ("PSP Load: Unable to create a new image\n");
	return -1;
    }

    if (image_type == INDEXED)
    {
	pspdword_t numcol, i, j;
	guchar * cmap;

	if (!PSPGetField (psp, PSPF_CMAPCOLORCOUNT, &numcol))
	{
	    g_message ("PSP Load: Unable to read colormap");
	    return -1;
	}

	cmap = g_new (guchar, 3 * numcol);

	for (i = 0, j = 0; i < numcol; i++)
	{
	    if (!PSPGetColor (psp, i, &cmap[j++], &cmap[j++], &cmap[j++]))
	    {
		g_message ("PSP Load: Unable to read colormap");
		return -1;
	    }
	    printf ("colormap: %lu: %02x %02x %02x\n", i, cmap[j-3],
		    cmap[j-2], cmap[j-1]);
	}

	gimp_image_set_cmap (imageID, cmap, numcol);
	
	g_free (cmap);
    }

    gimp_image_set_filename (imageID, filename);

    	
#if GIMP_HAVE_UNITS
    if (PSPGetField (psp, PSPF_IMAGERESUNIT, &unit))
    {
	if ((unit >= 1) && (unit <= 2)
	    && PSPGetField (psp, PSPF_IMAGERESVALUE, &resolution))
	{
	    resolution *= psp2gimpunit[unit].factor;
		/* No separate X and Y resolutions in PSP */
#if VERBOSE
	    printf ("Resolution: %f per inch. Unit: %s\n", resolution,
		    (psp2gimpunit[unit].unit == UNIT_MM) ? "mm" :
		    (psp2gimpunit[unit].unit == UNIT_INCH ? "inch" :
		     "other"));
		
#endif
	    gimp_image_set_resolution (imageID, (float) resolution,
				       (float) resolution);
	    gimp_image_set_unit (imageID, psp2gimpunit[unit].unit);
	}
    }
#endif /* GIMP_HAVE_UNITS */
    
    tile_height = gimp_tile_height ();

    if (!PSPSetLayer (psp, PSP_LAYERFIRST))
    {
	g_message ("PSP Load: PSP file broken?");
	gimp_image_delete (imageID);
	return -1;
    }

    if (!PSPGetField (psp, PSPF_IMAGEACTIVELAYER, &activelayer))
	activelayer = -1;

#if VERBOSE
    printf ("Active Layer number: %ld\n", activelayer);
#endif
    
    i = 0;
#if GIMP_MAJOR_VERSION >= 1 && GIMP_MINOR_VERSION >= 1
    firstlinkgroup = 0;
#endif /* GIMP_VERSION >= 1.1.0 */
    do
    {
#if GIMP_MAJOR_VERSION >= 1 && GIMP_MINOR_VERSION >= 1
	int layertype;
#endif /* GIMP_VERSION >= 1.1.0 */	
	
	gimp_progress_update ((double) i / (double) layercount);
	if ((layerID = load_layer (psp, imageID, image_type)) == -1)
	{
	    gimp_image_delete (imageID);
	    return -1;
	}
	if (PSPGetField (psp, PSPF_LAYERNUMBER, &layernum))
	{
	    if (layernum == activelayer)
	    {
		activelayerID = layerID;
#if VERBOSE
		printf ("Active Layer ID: %d\n", layerID);
#endif
	    }
	}

#if GIMP_MAJOR_VERSION >= 1 && GIMP_MINOR_VERSION >= 1
	if (PSPGetField (psp, PSPF_LAYERLINKGROUP, &linkgroup))
	{
	    if (linkgroup > 0)
	    {
		if (firstlinkgroup == 0)
		    firstlinkgroup = linkgroup;
		if (firstlinkgroup == linkgroup)
		    gimp_layer_set_linked (layerID, 1);
	    }
	}
#endif /* GIMP_VERSION >= 1.1.0 */

#if GIMP_MAJOR_VERSION >= 1 && GIMP_MINOR_VERSION >= 1
	if (!PSPGetField (psp, PSPF_LAYERTYPE, &layertype))
	{
	    g_message ("PSP Load: Unable to get layer type");
	    return -1;
	}
	if (layertype == PSP_LAYER_TYPE_FLOATSEL)
	{
	    if (gimp_floating_sel_attach (layerID, layerID))
	    {
		g_message ("PSP Load: Unable to turn layer to floating");
	    }
	}
#endif /* GIMP_VERSION >= 1.1.0 */

	i++;
    } while (PSPSetLayer (psp, PSP_LAYERNEXT));

	/* Create the Selection if there is one (are some).
	 * At the moment, selections are simply converted to alpha
	 * channels, but this should change by calling the appropriate
	 * procedures to convert the channel to a selection directly,
	 * instead of letting the user do this by hand.
	 */
    if (PSPSetChannel (psp, PSP_CHANNELSELECTIONFIRST))
    {
	psprect_t extent;
	psprect_t dataextent;
	gint32 channel;
#define CHANNELNAMELEN 64
	char name [CHANNELNAMELEN];
	int i;
	guchar black[3] = { 0, 0, 0 };
	GDrawable * drawable;
	GPixelRgn pixel_rgn;
	unsigned long rows, cols;

#if VERBOSE
	printf ("Retrieving selection from image file\n");
#endif

	if (PSPGetField (psp, PSPF_SELECTIONEXTENT, &extent)
	    && PSPGetField (psp, PSPF_SELECTIONDATAEXTENT, &dataextent))
	{
	    cols = dataextent.xmax - dataextent.xmin;
	    rows = dataextent.ymax - dataextent.ymin;

	    i = 1;
	    do {
#if HAS_NPRINTF
		snprintf (name, CHANNELNAMELEN,
			  "Selection %i (auto-converted)", i);
#else
		sprintf (name, "Selection %i (auto-converted)", i);
#endif
#if VERBOSE
		printf ("gimp_channel_new (%d, \"%s\", %lu, %lu, %.1f, %lux"
			" [%d, %d, %d]);\n",
			imageID, name, width, height, (gdouble) 100.0,
			(unsigned long) black & 0xffffff,
			black[0], black[1], black[2]);
#endif
		if ((channel = gimp_channel_new (imageID, name, width, height,
						 (gdouble) 100.0,
						 black)) != -1)
		{
#if VERBOSE
		    printf ("Channel: name=%s id=%d\n", name, channel);
#endif    
		    gimp_channel_set_visible (channel, FALSE);
		    gimp_image_add_channel (imageID, channel, 0);
		    drawable = gimp_drawable_get (channel);

		    gimp_pixel_rgn_init (&pixel_rgn, drawable,
					 0, 0,
					 width, height, TRUE, FALSE);

		    set_channel (&pixel_rgn, width, height, drawable->bpp,
				 0, 0, 1);

		    if (!load_channel (psp, &pixel_rgn, cols, rows,
				       drawable->bpp, 0, 1))
		    {
			g_message ("PSP Load Error: Read of channel content "
				   "failed\n");
			gimp_image_delete (imageID);
			return -1;
		    }
		    gimp_drawable_flush (drawable);
		    gimp_drawable_detach (drawable);

#if GIMP_MAJOR_VERSION == 1 && GIMP_MINOR_VERSION == 0 /* Interface changed */
		    gimp_channel_ops_offset (imageID, channel, FALSE, 1,
					     extent.xmin + dataextent.xmin,
					     extent.ymin + dataextent.ymin);
#if 1 || SELECTION_LOAD
		    gimp_selection_load (imageID, channel);
#else
		    printf ("Selection not loaded!\n");
#endif
#else
		    gimp_channel_ops_offset (channel, FALSE, 1,
					     extent.xmin + dataextent.xmin,
					     extent.ymin + dataextent.ymin);
#if 1 || SELECTION_LOAD
		    gimp_selection_load (channel);
#else
		    printf ("Selection not loaded!\n");
#endif
#endif
#if 1 || SELECTION_LOAD
#if VERBOSE
		    printf ("gimp_image_remove_channel (imageID = %d,\n"
			    "                           channelID = %d);\n",
			    imageID, channel);
#endif
		    gimp_image_remove_channel (imageID, channel);
#endif
			/* Should ask about the next, but from the undo
			 * code I assume it is correct to not run
			 * gimp_channel_delete
			 */
#if 0 || SELECTION_LOAD
#if VERBOSE
		    printf ("gimp_channel_delete (channelID = %ld);\n",
			    channel);
#endif
		    gimp_channel_delete (channel);
#if VERBOSE
		    printf ("selection finished!\n");
#endif
#endif
		}
		else
		{
		    printf ("Channel creation for %s failed\n", name);
		    gimp_image_delete (imageID);
		    return -1;
		}
	    } while (PSPNextChannel (psp));
	}
    }

	/* Create the Alpha channels if there are some.
	 */
    if (PSPSetLayer (psp, PSP_LAYERFIRSTALPHA))
    {
	psprect_t extent;
	psprect_t dataextent;
	gint32 channel;
	char * name;
	int i;
	guchar black[3] = { 0, 0, 0 };
	GDrawable * drawable;
	GPixelRgn pixel_rgn;
	unsigned long rows, cols;

#if VERBOSE
	printf ("Importing alpha channel from image file\n");
#endif

	i = 1;
	do {
	    if (!PSPGetField (psp, PSPF_LAYERNAME, &name)
		|| !PSPGetField (psp, PSPF_LAYEREXTENT, &extent)
		|| !PSPGetField (psp, PSPF_LAYERDATAEXTENT, &dataextent)
		|| !PSPSetChannel (psp, PSP_CHANNELALPHAFIRST))
	    {
		g_message ("PSP Load: Unable to import alpha channel data");
		return -1;
	    }
	    
		/* Only one channel/bitmap per alpha channel supported */
		
	    cols = dataextent.xmax - dataextent.xmin;
	    rows = dataextent.ymax - dataextent.ymin;
	    
#if VERBOSE
		printf ("gimp_channel_new (%d, %s, %lu, %lu, %.1f, %lux"
			" [%d, %d, %d]);\n",
			imageID, name, width, height, (gdouble) 100.0,
			(unsigned long) black,
			black[0], black[1], black[2]);
#endif
	    if ((channel = gimp_channel_new (imageID, name, width, height,
					     (gdouble) 100.0,
					     black)) != -1)
	    {
#if VERBOSE
		printf ("Channel: name=%s id=%d\n", name, channel);
#endif
		PSPFree (name);

		gimp_channel_set_visible (channel, FALSE);
		gimp_image_add_channel (imageID, channel, 0);
		drawable = gimp_drawable_get (channel);

		gimp_pixel_rgn_init (&pixel_rgn, drawable,
				     0, 0,
				     width, height, TRUE, FALSE);
		    
		set_channel (&pixel_rgn, width, height, drawable->bpp,
			     0, 0, 1);

#if VERBOSE
		printf ("load_channel (PSP * psp = 0x%08lx,\n"
			"              GPixelRgn * pixel_rgn = 0x%08lx,\n"
			"              unsigned long cols = %lu,\n"
			"              unsigned long rows = %lu,\n"
			"              int bpp = %d,\n"
			"              int offset = %d,\n"
			"              int shift = %d);\n",
			(unsigned long) psp,
			(unsigned long) &pixel_rgn,
			cols, rows, drawable->bpp, 0, 1);
#endif
		if (!load_channel (psp, &pixel_rgn, cols, rows,
				   drawable->bpp, 0, 1))
		{
		      g_message ("PSP Load Error: Read of alpha channel "
				 "content failed\n");
		      return -1;
		}
		gimp_drawable_flush (drawable);
		gimp_drawable_detach (drawable);

#if GIMP_MAJOR_VERSION == 1 && GIMP_MINOR_VERSION == 0 /* Interface changed */
		    gimp_channel_ops_offset (imageID, channel, FALSE, 1,
					     extent.xmin + dataextent.xmin,
					     extent.ymin + dataextent.ymin);
#else
		    gimp_channel_ops_offset (channel, FALSE, 1,
					     extent.xmin + dataextent.xmin,
					     extent.ymin + dataextent.ymin);
#endif
	    }
	    else
	    {
		g_message ("Alpha channel creation failed\n");
		return -1;
	    }
		/* FIXME! Here, we could update the progress bar */
	} while (PSPNextLayer (psp));
    }

    if (activelayerID != -1)
	gimp_image_set_active_layer (imageID, activelayerID);

    PSPClose (psp);
    return imageID;
}

static gint
save_psp_image (char * filename) 
{
    PSP * psp;

    if ((psp = PSPOpen (filename, "wb")) == NULL)
	return 0;
    PSPClose (psp);

    return 1;
}

static void
savedialog_close_callback (GtkWidget * widget,
			   gpointer data)
{
    gtk_main_quit ();
}

static void
savedialog_ok_callback (GtkWidget * widget,
			gpointer data)
{
    psp_do_save = TRUE;
    gtk_widget_destroy (GTK_WIDGET (data));
}

#if 0
static void
savedialog_cancel_callback (GtkWidget * widget,
			    gpointer data)
{
    gtk_widget_destroy (GTK_WIDGET (data));
}
#endif

static void
savedialog_toggled_update (GtkWidget * widget,
			   gpointer data)
{
    if (GTK_TOGGLE_BUTTON (widget)->active)
	*(int*)data = TRUE;
    else
	*(int*)data = FALSE;
}

/****************************************************************
 * save_dialog() - Pop up the save dialog.
 ****************************************************************/

static gint
save_dialog(void)
{
    GtkWidget	*dialog,	/* Dialog window */
		*button,	/* OK/cancel buttons */
		*frame,		/* Frame for dialog */
		*table,		/* Table for dialog options */
		*vbox,		/* Vertical Box */
		*toggle,	/* Interlace toggle button */
		*label,		/* Label for controls */
		*notebook,
		*entry,
		*texttable,
		*scrollbar,
		*text;
    
    
    GSList	*group;		/* Radiobuttons group */
    gchar	**argv;		/* Fake command-line args */
    gint	argc;		/* Number of fake command-line args */
  
    int		compression_none = 1;
    int		compression_packbits = 0;
    int		compression_lz77 = 0;
  

    psp_do_save = FALSE;
  
      /* Fake the command-line args and read the rc... */

    argc    = 2;
    argv    = g_new (gchar *, 1);
    argv[0] = g_strdup("pspsave");
    argv[1] = g_strdup("--gtk-debug");

    gtk_init (&argc, &argv);
    gtk_rc_parse (gimp_gtkrc());

	/* Open a dialog window... */

    dialog = gtk_dialog_new();
    gtk_window_set_title (GTK_WINDOW(dialog), "Save as PSP");
    gtk_window_position (GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
    gtk_signal_connect (GTK_OBJECT(dialog), "destroy",
			(GtkSignalFunc) savedialog_close_callback, NULL);

	/* OK button... */

    button = gtk_button_new_with_label ("OK");
    GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
    gtk_signal_connect (GTK_OBJECT(button), "clicked",
			(GtkSignalFunc) savedialog_ok_callback,
			GTK_OBJECT(dialog));
    gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dialog)->action_area),
			button, TRUE, TRUE, 0);
    gtk_widget_grab_default(button);
    gtk_widget_show(button);

	/* Cancel button... */

    button = gtk_button_new_with_label ("Cancel");
    GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
    gtk_signal_connect_object (GTK_OBJECT(button), "clicked",
			       (GtkSignalFunc) gtk_widget_destroy,
			       GTK_OBJECT(dialog));
    gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dialog)->action_area),
			button, TRUE, TRUE, 0);
    gtk_widget_show (button);

	/* Options frame: */
    notebook = gtk_notebook_new ();
	/* This is already default */
    gtk_notebook_set_tab_pos (GTK_NOTEBOOK (notebook), GTK_POS_TOP);
    gtk_container_border_width (GTK_CONTAINER (notebook), 10);
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), notebook,
			TRUE, TRUE, 0);
    
	/* gtk_table_attach_defaults (GTK_TABLE (table), notebook, 0, 6, 0, 1);*/
    gtk_widget_show (notebook);

    vbox = gtk_vbox_new (FALSE, 0);
    frame = gtk_frame_new ("Compression");
    gtk_container_border_width (GTK_CONTAINER (frame), 10);
    gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 0);

/*    label = gtk_label_new ("Parameters Frame Label");
    gtk_container_add (GTK_CONTAINER (frame), label); 
    gtk_widget_show (label);
*/

	/* Compression setup */

    table = gtk_table_new (3, 1, FALSE);
/*    gtk_frame_set_shadow_type (GTK_FRAME (table),
      GTK_SHADOW_ETCHED_IN); */
    gtk_container_border_width (GTK_CONTAINER (table), 5);
    gtk_container_add (GTK_CONTAINER (frame), table);

	/* No Compression */
    group = NULL;
    toggle = gtk_radio_button_new_with_label (group, "None");
    group = gtk_radio_button_group (GTK_RADIO_BUTTON (toggle));
    gtk_table_attach (GTK_TABLE (table), toggle, 0, 1, 0, 1,
		      GTK_FILL | GTK_SHRINK | GTK_EXPAND, 0, 0, 0);
    gtk_signal_connect (GTK_OBJECT (toggle), "toggled",
			(GtkSignalFunc) savedialog_toggled_update,
			&compression_none);
#if 1
    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (toggle),
				 compression_none);
#else
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toggle),
				  compression_none);
#endif
    gtk_widget_show (toggle);

				  /* Packbits aka RLE */
    toggle = gtk_radio_button_new_with_label (group, "Pack-Bits (RLE)");
    group = gtk_radio_button_group (GTK_RADIO_BUTTON (toggle));
    gtk_table_attach (GTK_TABLE (table), toggle, 0, 1, 1, 2,
		      GTK_FILL | GTK_SHRINK | GTK_EXPAND, 0, 0, 0);
    gtk_signal_connect (GTK_OBJECT (toggle), "toggled",
			(GtkSignalFunc) savedialog_toggled_update,
			&compression_packbits);
#if 1
    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (toggle),
				 compression_packbits);
#else
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toggle),
				  compression_packbits);
#endif
    gtk_widget_show (toggle);

				  /* LZ77 compression */
    toggle = gtk_radio_button_new_with_label (group, "LZ77");
    group = gtk_radio_button_group (GTK_RADIO_BUTTON (toggle));
    gtk_widget_set_sensitive (toggle, FALSE);
    gtk_table_attach (GTK_TABLE (table), toggle, 0, 1, 2, 3,
		      GTK_FILL | GTK_SHRINK | GTK_EXPAND, 0, 0, 0);
    gtk_signal_connect (GTK_OBJECT (toggle), "toggled",
			(GtkSignalFunc) savedialog_toggled_update,
			&compression_lz77);
#if 1
    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (toggle),
				 compression_lz77);
#else
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toggle),
				  compression_lz77);
#endif
    gtk_widget_show (toggle);

    gtk_widget_show (table);
    gtk_widget_show (frame);
    frame = gtk_frame_new ("Note: File save not implemented yet!");
    gtk_container_border_width (GTK_CONTAINER (frame), 10);
    gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 0);

    table = gtk_table_new (1, 1, FALSE);
/*    gtk_frame_set_shadow_type (GTK_FRAME (table),
      GTK_SHADOW_ETCHED_IN); */
    gtk_container_border_width (GTK_CONTAINER (table), 5);
    gtk_container_add (GTK_CONTAINER (frame), table);
    
    label = gtk_label_new ("!!! The file you selected will be overwritten with practically useless data !!!");
    gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1, 0, 0, 0, 0);
    gtk_widget_show (label);

    gtk_widget_show (table);
    gtk_widget_show (frame);
    gtk_widget_show (vbox);

    label = gtk_label_new ("Parameters");
    gtk_notebook_append_page (GTK_NOTEBOOK (notebook), vbox, label);

    frame = gtk_frame_new ("Creator");
    gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_ETCHED_IN);
    gtk_container_border_width (GTK_CONTAINER (frame), 10);

    table = gtk_table_new (6, 2, FALSE);
    gtk_table_set_row_spacings (GTK_TABLE (table), 5);
    gtk_table_set_col_spacings (GTK_TABLE (table), 10);
/*    gtk_widget_set_usize (table, -1, -1); */

    label = gtk_label_new ("Title:");
    gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_RIGHT);
    gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1, 0, 0, 0, 0);
    gtk_widget_show (label);
    entry = gtk_entry_new ();
    gtk_entry_set_text (GTK_ENTRY (entry), "This is the title, it is very "
			"very long to test how gtk reacts if there is a "
			"big big widget contained within the window.  "
			"I just want to know if the full screen size "
			"would be used if no minimum size for the window "
			"is set here which would look rather strange.  "
			"On the other hand, what would happen if you a "
			"very big font is used.");
    gtk_table_attach (GTK_TABLE (table), entry, 1, 2, 0, 1,
		      GTK_FILL | GTK_SHRINK | GTK_EXPAND, 0, 0, 0);
    gtk_widget_show (entry);

    label = gtk_label_new ("Author:");
    gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_RIGHT);
    gtk_table_attach (GTK_TABLE (table), label, 0, 1, 1, 2, 0, 0, 0, 0);
    gtk_widget_show (label);
    entry = gtk_entry_new ();
    gtk_entry_set_text (GTK_ENTRY (entry), "I am the author");
    gtk_table_attach (GTK_TABLE (table), entry, 1, 2, 1, 2,
		      GTK_FILL | GTK_SHRINK | GTK_EXPAND, 0, 0, 0);
    gtk_widget_show (entry);

    label = gtk_label_new ("Copyright:");
    gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_RIGHT);
    gtk_table_attach (GTK_TABLE (table), label, 0, 1, 2, 3, 0, 0, 0, 0);
    gtk_widget_show (label);
    entry = gtk_entry_new ();
    gtk_entry_set_text (GTK_ENTRY (entry), " 1999 by me");
    gtk_table_attach (GTK_TABLE (table), entry, 1, 2, 2, 3,
		      GTK_FILL | GTK_SHRINK | GTK_EXPAND, 0, 0, 0);
    gtk_widget_show (entry);

    label = gtk_label_new ("Description:");
    gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_RIGHT);
    gtk_table_attach (GTK_TABLE (table), label, 0, 1, 3, 4, 0, 0, 0, 0);
    gtk_widget_show (label);

    texttable = gtk_table_new (2, 1, FALSE);
    text = gtk_text_new (NULL, NULL);
    gtk_widget_set_usize (text, -1, 40);
	/* Put the string with default look in the area */
	/* enable automatic word-wrap in the text area */
    gtk_text_set_word_wrap (GTK_TEXT (text), TRUE);
    gtk_text_set_editable (GTK_TEXT (text), TRUE);
    gtk_table_attach (GTK_TABLE (texttable), text, 0, 1, 0, 1,
		      GTK_FILL | GTK_SHRINK | GTK_EXPAND,
		      GTK_FILL | GTK_SHRINK | GTK_EXPAND, 0, 0);
    scrollbar = gtk_vscrollbar_new (GTK_TEXT (text)->vadj);
    gtk_table_attach (GTK_TABLE (texttable), scrollbar, 1, 2, 0, 1,
		      GTK_FILL,
		      GTK_FILL | GTK_SHRINK | GTK_EXPAND, 0, 0);
    gtk_widget_show (scrollbar);
    gtk_widget_show (text);
    gtk_table_attach (GTK_TABLE (table), texttable, 1, 2, 3, 4,
		      GTK_FILL | GTK_SHRINK | GTK_EXPAND,
		      GTK_FILL | GTK_SHRINK | GTK_EXPAND, 0, 0);
    gtk_widget_show (texttable);

    label = gtk_label_new ("Creation date:");
    gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_RIGHT);
    gtk_table_attach (GTK_TABLE (table), label, 0, 1, 4, 5, 0, 0, 0, 0);
    gtk_widget_show (label);
    entry = gtk_entry_new ();
    gtk_entry_set_text (GTK_ENTRY (entry), "Sun Feb 21 10:20:42 CET 1999");
    gtk_table_attach (GTK_TABLE (table), entry, 1, 2, 4, 5,
		      GTK_FILL | GTK_SHRINK | GTK_EXPAND, 0, 0, 0);
    gtk_widget_show (entry);

    label = gtk_label_new ("Application:");
    gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_RIGHT);
    gtk_table_attach (GTK_TABLE (table), label, 0, 1, 5, 6, 0, 0, 0, 0);
    gtk_widget_show (label);
    entry = gtk_entry_new ();
    gtk_entry_set_text (GTK_ENTRY (entry), "Gimp 1.0.2.0");
    gtk_entry_set_editable (GTK_ENTRY (entry), FALSE);
    gtk_table_attach (GTK_TABLE (table), entry, 1, 2, 5, 6,
		      GTK_FILL | GTK_SHRINK | GTK_EXPAND, 0, 0, 0);
    gtk_widget_show (entry);

    gtk_container_border_width (GTK_CONTAINER (table), 10);
    gtk_container_add (GTK_CONTAINER (frame), table);
    
    gtk_widget_show (table);

    label = gtk_label_new ("Image Information");
    gtk_notebook_append_page (GTK_NOTEBOOK (notebook), frame, label);

    gtk_widget_show (frame);

#if 1
    gtk_widget_realize (text);
    gtk_text_freeze (GTK_TEXT (text));
    gtk_text_insert (GTK_TEXT (text), NULL, NULL, NULL,
		     "I am the description"
		     "used to describe the image"
		     "similar to a gif comment and more", -1);
    gtk_text_thaw (GTK_TEXT (text));
#endif

    gtk_widget_show (dialog);

    gtk_main();
    gdk_flush();

    return (psp_do_save);
}
